# @Author: carlosgilgonzalez
# @Date:   2019-08-16T00:57:12+01:00
# @Last modified by:   carlosgilgonzalez
# @Last modified time: 2019-10-21T01:01:29+01:00

"""
Python lib to interface with micropython devices through WebREPL protocol or
through Serial connection.

Example usage:

WIRELESS DEVICE (WebREPL)
    >>> from upydevice import W_UPYDEVICE
# Setup and configurate a device :
    >>> esp32 = W_UPYDEVICE('192.168.1.56', 'mypass') # (target_ip, password)
# Send command:
    >>> esp32.cmd('led.on()')
    >>> esp32.cmd("uos.listdir('/')")
    ['boot.py', 'webrepl_cfg.py', 'main.py'] # this output is stored in [upydevice].output
    >>> esp32.output
    ['boot.py', 'webrepl_cfg.py', 'main.py']
    >>> esp32.cmd('foo()')
    >>> esp32.cmd('x = [1,2,3];my_var = len(x);print(my_var)')
    3
# Soft Reset:
    >>> esp32.reset()
    Rebooting device...

    ### closed ###

    Done!

SERIAL DEVICE (Picocom, Pyserial)
    >>> from upydevice import S_UPYDEVICE
# Setup and configurate a device :
    >>> esp32 = S_UPYDEVICE('/dev/tty.SLAB_USBtoUART', 1000, 115200) # defaults (serial_port, timeout=1000, baudrate=9600)
# Send command:
    >>> esp32.cmd('led.on()')
    >>> esp32.cmd("uos.listdir('/')")
    ['boot.py', 'webrepl_cfg.py', 'main.py'] # this output is stored in [upydevice].output
    >>> esp32.output
    ['boot.py', 'webrepl_cfg.py', 'main.py']
    >>> esp32.cmd('foo()')
    >>> esp32.cmd('x = [1,2,3];my_var = len(x);print(my_var)')
    3
# Soft Reset:
    >>> esp32.reset()
    Rebooting device...
    Done!

PYBOARD (Picocom, Pyserial)
    >>> from upydevice import PYBOARD
# Setup and configurate a device :
    pyboard = PYBOARD('/dev/tty.usbmodem3370377430372') # defaults (serial_port, timeout=1000, baudrate=9600)
# Send command:
    >>> pyboard.cmd('pyb.LED(1).toggle()',timeout=100)
    >>> pyboard.cmd("import uos;uos.listdir('/flash')")
    ['main.py', 'pybcdc.inf', 'README.txt', 'boot.py', '.fseventsd', '.Trashes'] # this output is stored in [upydevice].output
    >>> pyboard.output
    ['main.py', 'pybcdc.inf', 'README.txt', 'boot.py', '.fseventsd', '.Trashes']
    >>> pyboard.cmd('foo()')
    >>> pyboard.cmd('x = [1,2,3];my_var = len(x);print(my_var)')
    3
# Soft Reset:
    >>> pyboard.reset()
    Rebooting pyboard...
    Done!

GROUP (to send commands to several devices at a time)
# Setup and configurate the devices :
    >>> from upydevice import W_UPYDEVICE, PYBOARD, GROUP
# PYBOARD
    >>> pyboard = PYBOARD('/dev/tty.usbmodem387E386731342')
# ESP32
    >>> esp32_A = W_UPYDEVICE('192.168.1.53', 'mypass')
    >>> esp32_B = W_UPYDEVICE('192.168.1.40', 'mypass')

# Setup and configurate the group:
    >>> my_group = GROUP([esp32_A, esp32_B, pyboard])

# Each upydevice has a name attribute that can be set at creation momment or after
# pyboard = PYBOARD('/dev/tty.usbmodem387E386731342', name='my_pyboard_1'); or pyboard.name = 'my_pyboard_1')
# If not set an automatic name will be set as 'upydev_class'+'ip or serial port'

# Send command:
    >>> my_group.cmd('import machine;import ubinascii;ubinascii.hexlify(machine.unique_id())')
    Sending command to wupydev_53
    b'30aea4233564'

    Sending command to wupydev_40
    b'807d3a809b30'

    Sending command to pyboard_tty.usbmodem387E386731342
    b'33004e000351343134383038'

# There is an option to silent the group messages with group_silent = True, and or each device ouput with device_silent=True

# Output is stored in group output attribute:
    >>> my_group.output
    {'wupydev_53': b'30aea4233564', 'wupydev_40': b'807d3a809b30', 'pyboard_tty.usbmodem387E386731342': b'33004e000351343134383038'}

# Send command parallel mode **: (experimental mode, may not work 100% of the times, depends on the connection quality (for wireless devices))
    >>> my_group.cmd_p('6*12')
    Sending command to: wupydev_53, wupydev_40, pyboard_tty.usbmodem387E386731342
    72


    72

    72

    Done!
# To see which ouput corresponds to which device use 'id=True' parameter:

    >>> my_group.cmd_p('ubinascii.hexlify(machine.unique_id())', id=True)
    Sending command to: wupydev_53, wupydev_40, pyboard_tty.usbmodem387E386731342
    pyboard_tty.usbmodem387E386731342:b'33004e000351343134383038'
    pyboard_tty.usbmodem387E386731342:
    pyboard_tty.usbmodem387E386731342:
    wupydev_40:b'807d3a809b30'
    wupydev_53:b'30aea4233564'
    wupydev_40:
    wupydev_53:
    Done!
    >>>
    >>> my_group.output
    {'wupydev_53': b'30aea4233564', 'wupydev_40': b'807d3a809b30', 'pyboard_tty.usbmodem387E386731342': b'33004e000351343134383038'}

To see more info read the DOCS at the github repo.
"""

from .upydevice import *
name = 'upydevice'



#!/usr/bin/env python3
# @Author: carlosgilgonzalez
# @Date:   2019-10-30T23:49:59+00:00
# @Last modified by:   carlosgilgonzalez
# @Last modified time: 2019-11-16T22:41:08+00:00


"""Phantom classes collection"""

from .upydevice import (upy_cmd_c_r, upy_cmd_c_raw_r, upy_cmd_c_r_in_callback,
                        upy_cmd_c_r_nb, upy_cmd_c_r_nb_in_callback)
import time
import socket
import struct
import sys
from datetime import datetime
from binascii import hexlify
import json
import os


# MICROPYTHON DEFAULT CLASSES

class MACHINE:
    def __init__(self, device):
        """Phantom MACHINE class"""
        self.d = device
        self.name = 'machine'
        self.dev_dict = {'name': self.name, 'dev': device}

    @upy_cmd_c_r()
    def unique_id(self):
        return self.dev_dict


# PIN

class Pin:
    def __init__(self, device, name, mode='OUT', number=0, init=False):
        """Phantom machine.Pin class"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}
        # CONST
        self.IRQ_FALLING = 2
        self.IRQ_RISING = 1
        self.IRQ_BOTH = 3
        self.OUT = 3
        self.IN = 1
        self.OPEN_DRAIN = 7
        self.PULL_HOLD = 4
        self.PULL_DOWN = 1
        self.WAKE_LOW = 4
        self.WAKE_HIGH = 5
        if init:
            if mode == 'IN':
                self.d.cmd("{}=Pin({}, {})".format(
                    name, number, self.IN), silent=True)
            if mode == 'OUT':
                self.d.cmd("{}=Pin({}, {})".format(
                    name, number, self.OUT), silent=True)

    @upy_cmd_c_r()
    def value(self, *args):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def on(self):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def off(self):
        return self.dev_dict

    def toggle(self):
        self.value(not self.value())

    @upy_cmd_c_r(rtn=False)
    def irq(self, trigger, handler, priority=1, wake=None, hard=False):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def mode(self, mode):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def pull(self, pull):
        return self.dev_dict


# I2C

class I2C:
    def __init__(self, device, name='i2c', init=False, scl=None, sda=None, freq=400000):
        """Phantom I2C class"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}
        if init:
            self.d.cmd('from machine import I2C,Pin; {} = I2C(scl=Pin({}), sda=Pin({}), freq={})'.format(
                name, scl, sda, freq), silent=True)

    @upy_cmd_c_r()
    def scan(self):
        return self.dev_dict

    # Primitive I2C operations (on I2C software only)

    @upy_cmd_c_r(rtn=False)
    def start(self):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def stop(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def readinto(self, buf, nack=True):
        return self.dev_dict

    @upy_cmd_c_r()
    def write(self, buf):
        return self.dev_dict

    # Standard bus operations

    @upy_cmd_c_r()
    def readfrom(self, addr, nbytes, stop=True):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def readfrom_into(self, addr, buf, stop=True):
        return self.dev_dict

    @upy_cmd_c_r()
    def writeto(self, addr, buf, stop=True):
        return self.dev_dict

    @upy_cmd_c_r()
    def writevto(self, addr, vector, stop=True):
        return self.dev_dict

    # Memory operations
    @upy_cmd_c_r()
    def readfrom_mem(self, addr, memaddr, nbytes, addrsize=8):
        return self.dev_dict

    @upy_cmd_c_r()
    def readfrom_mem_into(self, addr, memaddr, buf, addrsize=8):
        return self.dev_dict

    @upy_cmd_c_r()
    def writeto_mem(self, addr, memaddr, buf, addrsize=8):
        return self.dev_dict


# UOS

class UOS:
    def __init__(self, device, name='uos'):
        """Phantom UOS class"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}

    @upy_cmd_c_r()
    def listdir(self, directory):
        return self.dev_dict

    @upy_cmd_c_r()
    def getcwd(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def mkdir(self, directory):
        return self.dev_dict

    @upy_cmd_c_r()
    def rmdir(self, directory):
        return self.dev_dict

    @upy_cmd_c_r()
    def chdir(self, directory):
        return self.dev_dict

    @upy_cmd_c_r()
    def remove(self, file):
        return self.dev_dict

    @upy_cmd_c_raw_r()
    def uname(self):
        return self.dev_dict

#############################################


# PYBOARD CLASSES

class pyb_LED:
    def __init__(self, device, number):
        """Phantom pyb.LED class"""
        self.d = device
        self.name = "{}({})".format('LED', number)
        self.dev_dict = {'name': self.name, 'dev': device}
        self.dev_dict['dev'].cmd('from pyb import LED', silent=True)

    @upy_cmd_c_r(rtn=False)
    def toggle(self):
        return self.dev_dict


class pyb_Timer:
    def __init__(self, device, name):
        """Phantom pyb Timer"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}

    @upy_cmd_c_r(rtn=False)
    def init(self, freq, callback):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def deinit(self):
        return self.dev_dict


class pyb_Servo:
    def __init__(self, device, name, number=None, init=False):
        """Phantom pyb Servo"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}
        if init:
            self.d.cmd('import pyb; {} = pyb.Servo({})'.format(name, number))

    @upy_cmd_c_r()
    def angle(self, *args, **kargs):  # to allow no args
        """args: angle, time"""
        return self.dev_dict

    @upy_cmd_c_r()
    def speed(self, *args, **kargs):  # to allow no args
        """args: speed, time"""
        return self.dev_dict

    @upy_cmd_c_r()
    def pulse_width(self, *args, **kargs):  # to allow no args
        """args: value"""
        return self.dev_dict

    @upy_cmd_c_r()
    def calibration(self, *args, **kargs):  # to allow no args
        """args: pulse_min, pulse_max, pulse_centre,[pulse_angle_90, pulse_speed_100]"""
        return self.dev_dict


#############################################


# TIMER

class machine_Timer:
    def __init__(self, device, name):
        """Phantom machine Timer"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}
        self.PERIODIC = 1
        self.ONE_SHOT = 0

    @upy_cmd_c_r(rtn=False)
    def init(self, mode, period, callback):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def deinit(self):
        return self.dev_dict


# NETWORK WLAN

class WLAN:
    def __init__(self, device, name='wlan'):
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}
        self.AUTHMODE_DICT = {0: 'NONE', 1: 'WEP',
                              2: 'WPA PSK', 3: 'WPA2 PSK', 4: 'WPA/WAP2 PSK'}
        self.net_ifconfig = None

    @upy_cmd_c_r()
    def scan(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def ifconfig(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def config(self, *args, **kargs):
        return self.dev_dict

    @upy_cmd_c_r()
    def status(self, rssi):
        return self.dev_dict

    def pty_scan(self):
        netscan_list = self.scan()
        print('=' * 110)
        print('{0:^20} | {1:^25} | {2:^10} | {3:^15} | {4:^15} | {5:^10} '.format(
            'ESSID', 'BSSID', 'CHANNEL', 'RSSI (dB)', 'AUTHMODE', 'HIDDEN'))
        print('=' * 110)
        for netscan in netscan_list:
            auth = self.AUTHMODE_DICT[netscan[4]]
            vals = hexlify(netscan[1]).decode()
            bssid = ':'.join([vals[i:i+2] for i in range(0, len(vals), 2)])
            print('{0:^20} | {1:^25} | {2:^10} | {3:^15} | {4:^15} | {5:^10} '.format(
                netscan[0].decode(), bssid, netscan[2], netscan[3],
                auth, str(netscan[5])))

    def get_ifconfig(self):

        ifconf = self.ifconfig()
        self.net_ifconfig = {
            'IP': ifconf[0], 'SUBNET': ifconf[1], 'GATEAWAY': ifconf[2],
            'DNS': ifconf[3]}
        return self.net_ifconfig

    def get_rssi(self):
        return self.status('rssi')


# NETWORK AP

class AP:
    def __init__(self, device, name='ap'):
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}
        self.AUTHMODE_DICT = {0: 'NONE', 1: 'WEP',
                              2: 'WPA PSK', 3: 'WPA2 PSK', 4: 'WPA/WAP2 PSK'}
        self.ap_ifconfig = None
        self.essid = None
        self.conn_devs = None

    @upy_cmd_c_r()
    def status(self, stats):
        return self.dev_dict

    @upy_cmd_c_r()
    def ifconfig(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def config(self, *args, **kargs):
        return self.dev_dict

    def get_ifconfig(self):

        ifconf = self.ifconfig()
        self.ap_ifconfig = {
            'IP': ifconf[0], 'SUBNET': ifconf[1], 'GATEAWAY': ifconf[2],
            'DNS': ifconf[3]}
        return self.ap_ifconfig

    def get_essid(self):

        self.essid = self.config('essid')
        return self.essid

    def get_scandevs(self, verbose=True):  # NOT COMPATIBLE WITH ESP8266
        self.conn_devs = self.status('stations')
        if verbose:
            print('Found {} devices'.format(len(self.conn_devs)))
        mac_addr = []
        for dev in self.conn_devs:
            bytdev = bytearray(dev[0])
            mac_ad = ':'.join(str(val) for val in bytdev)
            mac_addr.append(mac_ad)
            if verbose:
                try:
                    print(f'MAC address: {mac_ad}')
                except Exception as e:
                    print("MAC address: {}".format(mac_ad))
                    pass

        return mac_addr

# NETWORK SOCKETS
# U_SOCKET

#############################################
# CUSTOM CLASSES (UPYDEVICE UTILS)


# IRQ UTILS

class IRQ_MG:
    def __init__(self, device, name, init_soc=False, port=8005, p_format='f',
                 n_vars=3, sensor=None, log_dir=None, logg=None):
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}
        self.soc = None
        self.port = port
        if init_soc:
            self.soc = socket_server(port=self.port)
        self.p_format = p_format
        self.n_vars = n_vars
        self.data_length = struct.calcsize(self.p_format*self.n_vars)
        self.buffer = []
        self.log_dir = log_dir
        # self.sensor, class U_IMU_IRQ (read_data, set_mode)

    @upy_cmd_c_r(rtn=False)
    def reset_flag(self):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def reset_flag_counter(self):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def set_irq_msg(self, msg):
        return self.dev_dict

    @upy_cmd_c_r()
    def irq_state(self):
        return self.dev_dict

    @upy_cmd_c_r(debug=True)  # to check prints
    def wait_irq(self, reset=True):
        return self.dev_dict

    @upy_cmd_c_r_nb(debug=True)  # to check prints
    def wait_async_irq(self, reset=True):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def buzz_beep(self, sleeptime, ntimes, ntimespaced, fq):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def led_blink(self, sleeptime, ntimes, ntimespaced):
        return self.dev_dict

    @upy_cmd_c_r_in_callback(rtn=False)
    def active_button(self, callback, **kargs):
        return self.dev_dict

    @upy_cmd_c_r_in_callback(rtn=False)
    def active_button_rev(self, callback, **kargs):
        return self.dev_dict

    def async_irq_listener_loop(self, on_irq, func_loop=None, wait_time=0.05):
        self.wait_async_irq()
        self.d.output = None
        while self.d.output is None:
            if func_loop is not None:
                func_loop()
            self.d.get_opt()
            time.sleep(wait_time)

        # check condition in self.d.output
        on_irq(self.d.output)

    # THIS CAN HANG WEBREPL (BETTER USE SOCKETS) --> continuos_async_soc...
    def continuos_async_irq_loop(self, on_irq, func_loop=None, wait_time=0.05):
        while True:
            try:
                self.async_irq_listener_loop(on_irq, func_loop,
                                             wait_time=wait_time)
                time.sleep(1)
            except KeyboardInterrupt:
                break

    def async_irq_listener_check(self, on_irq, waiting=True):
        if not waiting:
            self.wait_async_irq()
            self.d.output = None
        if self.d.output is None:
            self.d.get_opt()
            if self.d.output is None:
                return False
            else:
                # check condition in self.d.output
                on_irq(self.d.output)
                return self.d.output
        else:
            # check condition in self.d.output
            on_irq(self.d.output)
            return self.d.output

    # SOCKETS MESSAGING

    @upy_cmd_c_r_nb(debug=True)
    def connect_SOC(self, host, port):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def disconnect_SOC(self):
        return self.dev_dict

    def start_server(self):
        self.connect_SOC(self.soc.host, self.port)
        self.soc.start_SOC()

    def stop_server(self):
        self.disconnect_SOC()
        self.soc.conn.close()
        self.soc.serv_soc.close()

    def soc_recv_message(self):
        try:
            data = self.soc.conn.recv(self.data_length)
            data_unpack = struct.unpack(self.p_format*self.n_vars, data)
            self.d.output = data_unpack
            return data_unpack
        except Exception as e:
            return None

    # + LOG_OPTION
    def async_soc_irq_listener_loop(self, on_irq, func_loop=None,
                                    wait_time=0.05, log_nf=False,
                                    buffer=False, log=False,
                                    filename=None, tag=None):
        if log_nf:
            log = True
            if filename is None:
                filename = 'irq_{}_log.txt'.format(self.name)
            self.lognow_shot('irq', filename=filename, rtn=False)
        self.d.output = None
        while self.d.output is None:
            if func_loop is not None:
                func_loop()
            self.d.output = self.soc_recv_message()
            time.sleep(wait_time)

        # check condition in self.d.output
        on_irq(self.d.output)   # PRINT + LOG_OPTION + BUFFER
        if log:
            self.log_data_shot(filename, self.d.output, n_tag=tag)
        if buffer:
            self.log_data_shot_buff(self.d.output, n_tag=tag)

    # + LOG_OPTION
    def continuos_async_soc_irq_loop(self, on_irq, func_loop=None,
                                     wait_time=0.05, filename=None,
                                     log=False, tag=None, buffer=False):
        if log:
            if filename is None:
                filename = 'irq_{}_log.txt'.format(self.name)
            self.lognow_shot('irq', filename=filename, rtn=False)
        while True:
            try:
                self.async_soc_irq_listener_loop(
                    on_irq, func_loop, wait_time=wait_time, log=log,
                    filename=filename, tag=tag, buffer=buffer)
                time.sleep(wait_time)
            except KeyboardInterrupt:
                break

    # CALLBACKS

    def buzzer_callback(self, x):
        pass

    def buzzer_callback_rev(self, x):
        pass

    def led_callback(self, x):
        pass

    def led_callback_rev(self, x):
        pass

    def msg_callback(self, x):
        pass

    def msg_callback_rev(self, x):
        pass

    def sensor_callback(self, x):
        pass

    def sensor_soc_callback(self, x):
        pass

    def data_print(self, x):
        try:
            print(self.data_print_msg.format(*x, self.header['UNIT']))
        except Exception as e:
            pass

    # LOG methods

    # FILE LOG

    def lognow_shot(self, sensor_mode, filename=None, debug=True, rtn=True):
        file_name = filename
        if debug:
            print('Saving file {} ...'.format(file_name))
        with open(file_name, 'w') as file_log:
            header_vars = self.header['VAR'].copy()
            header_vars.append('TS')
            header_unit = self.header['UNIT']
            SHOT_HEADER = {'VAR': header_vars, 'UNIT': header_unit}
            file_log.write(json.dumps(SHOT_HEADER))
            file_log.write('\n')
        if rtn:
            return file_name

    def log_data_shot(self, filename, data, n_tag=None):
        try:
            data_pack = dict(zip(self.header['VAR']+['TS'],
                                 [val for val in data]+[datetime.now().strftime("%m_%d_%Y_%H_%M_%S")]))
            if n_tag is not None:
                data_pack = dict(zip(self.header['VAR']+['TS'],
                                     [val for val in data]+[n_tag]))
            with open(filename, 'a') as file_log:
                file_log.write(json.dumps(data_pack))
                file_log.write('\n')
        except Exception as e:
            pass

    # BUFFER LOG

    def log_data_shot_buff(self, data, n_tag=None):
        try:
            data_pack = dict(zip(self.header['VAR']+['TS'],
                                 [val for val in data]+[datetime.now().strftime("%m_%d_%Y_%H_%M_%S")]))
            if n_tag is not None:
                data_pack = dict(zip(self.header['VAR']+['TS'],
                                     [val for val in data]+[n_tag]))
            self.buffer.append(json.dumps(data_pack))
        except Exception as e:
            pass

    def read_buffer_shot(self):
        list_of_vars = {var: [] for var in self.header['VAR']+['TS']}
        for message in self.buffer:
            try:
                vals = json.loads(message)
                for key in vals.keys():
                    list_of_vars[key].append(vals[key])
            except Exception as e:
                pass
        return (list_of_vars)

    def flush_buffer(self):
        self.buffer = []


# TCP STREAMER
class STREAMER:
    def __init__(self, device, name, init_soc=False, port=8005, p_format='f',
                 n_vars=3, log_dir=None, chunk_buffer_size=20, soc_timeout=1,
                 logg=None):
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}
        self.soc = None
        self.port = port
        if init_soc:
            self.soc = socket_server(port=self.port, soc_timeout=soc_timeout,
                                     logg=logg)
        self.p_format = p_format
        self.n_vars = n_vars
        self.data_length = struct.calcsize(self.p_format*self.n_vars)
        self.chunk_buffer_size = chunk_buffer_size
        self.chunk_buffer_data_length = struct.calcsize(
            self.p_format*self.chunk_buffer_size)
        self.fq = None
        self.log_dir = log_dir
        self.buffer = []
        self.time_test = 0
        self._json_errors = 0
        self._json_buffer = ' '

    # STREAM CLASS INHERITANCE
    @upy_cmd_c_r()
    def read_data(self):
        return self.dev_dict

    # SOCKETS METHODS
    @upy_cmd_c_r_nb(debug=True)
    def connect_SOC(self, host, port):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def disconnect_SOC(self):
        return self.dev_dict

    def start_server(self):
        self.connect_SOC(self.soc.host, self.port)
        self.soc.start_SOC()

    def stop_server(self):
        self.disconnect_SOC()
        self.soc.conn.close()
        self.soc.serv_soc.close()

    def soc_recv_message(self):
        try:
            data = self.soc.conn.recv(self.data_length)
            data_unpack = struct.unpack(self.p_format*self.n_vars, data)
            self.d.output = data_unpack
            return data_unpack
        except Exception as e:
            return None

    def soc_recv_chunk_message(self):
        try:
            data = self.soc.conn.recv(self.chunk_buffer_data_length)
            data_unpack = struct.unpack(self.p_format*self.chunk_buffer_size,
                                        data)
            self.d.output = data_unpack
            return data_unpack
        except Exception as e:
            return None

    # JSON

    def soc_recv_chunk_message_json(self):
        try:
            data = b' '
            while data.decode() != '}':
                data = self.soc.conn.recv(1)
                self._json_buffer += data.decode()
            data = b' '
            data_unpack = json.loads(self._json_buffer)
            self._json_buffer = ' '
            return data_unpack
        except Exception as e:
            print(e)
            self._json_errors += 1
            return None

    def is_chunk(self, x):
        if len(x) > self.n_vars:
            return True
        else:
            return False

    # STREAM METHODS
    @upy_cmd_c_r_nb(rtn=False)
    def sample_send(self):
        return self.dev_dict

    def soc_read_data(self, timeout=0.2):  # ONE SHOT READ (+ LOG_OPTION)
        self.sample_send()
        time.sleep(timeout)
        return self.soc_recv_message()

    def shot_read(self, on_message, timeout=0.5, log=False, name_file=None,
                  tag=None, buffer=False):
        soc_data = self.soc_read_data(timeout=timeout)
        on_message(soc_data)
        if log:
            if name_file not in os.listdir(self.log_dir):
                self.lognow_shot(self.sens_mode, filename=name_file, debug=True,
                                 rtn=False)

                self.log_data_shot(name_file, soc_data, n_tag=tag)
            else:
                self.log_data_shot(name_file, soc_data, n_tag=tag)
        if buffer:
            self.log_data_shot_buff(soc_data, n_tag=tag)

    @upy_cmd_c_r_nb_in_callback(rtn=False)
    def start_send(self, sampling_callback, timeout=100, on_init=None):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def stop_send(self):
        return self.dev_dict

    # CONTINUOS STREAM (+ LOG_OPTION (FILE OR BUFFER)) +(TEST OPTION)
    def continuous_stream(self, on_message, timeout=100, init=True,
                          static=False, log=False, buffer=False,
                          on_init=None, test=False):
        if init:
            self.fq = 1/(timeout/1000)
            self.header['fq(hz)'] = self.fq
            self.start_send(
                sampling_callback=self.sample_send_call, timeout=timeout, on_init=on_init)

        if log:
                name_file = self.lognow(self.sens_mode)
        if test:
            t0 = time.time()
            buffer = True
        if not static:
            while True:
                try:
                    soc_data = self.soc_recv_message()
                    on_message(soc_data)
                    if log:
                        self.log_data(name_file, soc_data)
                    if buffer:
                        self.log_data_buff(soc_data)
                except KeyboardInterrupt:
                    if test:
                        self.time_test = abs(time.time()-t0)
                    self.stop_send()
                    print('\n')
                    self.soc.flush()
                    print('Done!')
                    break
        else:
            print(self.header_msg.format(self.sens_mode, self.header['UNIT'],
                                         self.fq))
            print('\n')
            print(('{:^15}'*len(self.header['VAR'])
                   ).format(*self.header['VAR']))
            while True:
                try:
                    soc_data = self.soc_recv_message()
                    on_message(soc_data)
                    if log:
                        self.log_data(name_file, soc_data)
                    if buffer:
                        self.log_data_buff(soc_data)
                except KeyboardInterrupt:
                    if test:
                        self.time_test = abs(time.time()-t0)
                    self.stop_send()
                    print('\n')
                    self.soc.flush()
                    print('Done!')
                    break

    def data_print(self, x):
        try:
            print(self.data_print_msg.format(*x, self.header['UNIT']))
        except Exception as e:
            pass

    def data_print_static(self, x):
        try:
            decode = dict(
                zip(self.header['VAR'], [format(val, '.4f') for val in x]))
            sys.stdout.write("\033[K")
            print(('{:^15}'*len(decode.values())
                   ).format(*decode.values()), end='\r')
            sys.stdout.flush()
        except Exception as e:
            pass

    def data_print_chunk_static(self, x):
        try:
            decode = dict(
                zip(self.header['VAR'], [format(val, '.4f') for val in x[0]]))
            sys.stdout.write("\033[K")
            print(('{:^15}'*len(decode.values())
                   ).format(*decode.values()), end='\r')
            sys.stdout.flush()
        except Exception as e:
            pass

    def data_print_chunk_static_json(self, x):
        try:
            decode = dict(
                zip(self.header['VAR'], [format(val, '.4f') for val in [x[key][0] for key in x.keys()]]))
            sys.stdout.write("\033[K")
            print(('{:^15}'*len(decode.values())
                   ).format(*decode.values()), end='\r')
            sys.stdout.flush()
        except Exception as e:
            pass

    def continuous_chunk_stream(self, timeout=100, init=True,
                               log=False, buffer=False, on_init=None,
                               test=False, static=False):
        if init:
            self.fq = 1/(timeout/1000)
            self.header['fq(hz)'] = self.fq
            self.start_send(
                sampling_callback=self.chunk_send_call, timeout=timeout, on_init=on_init)

        if log:
                name_file = self.lognow(self.sens_mode)
        if test:
            t0 = time.time()
            buffer = True
        if not static:
            while True:
                try:
                    soc_data = self.soc_recv_chunk_message()
                    if log:
                        self.log_data_chunk(name_file, soc_data)
                    if buffer:
                        self.log_data_chunk_buff(soc_data)
                except KeyboardInterrupt:
                    if test:
                        self.time_test = abs(time.time()-t0)
                    self.stop_send()
                    print('\n')
                    self.soc.flush()
                    print('Done!')
                    break
        else:
            print(self.header_msg.format(self.sens_mode, self.header['UNIT'],
                                         self.fq))
            print('\n')
            print(('{:^15}'*len(self.header['VAR'])
                   ).format(*self.header['VAR']))
            while True:
                try:
                    soc_data = self.soc_recv_chunk_message()
                    self.data_print_chunk_static(soc_data)
                    if log:
                        self.log_data_chunk(name_file, soc_data)
                    if buffer:
                        self.log_data_chunk_buff(soc_data)
                except KeyboardInterrupt:
                    if test:
                        self.time_test = abs(time.time()-t0)
                    self.stop_send()
                    print('\n')
                    self.soc.flush()
                    print('Done!')
                    break

    def continuous_chunk_stream_json(self, timeout=100, init=True,
                                    log=False, buffer=False, on_init=None,
                                    test=False, static=False):
        if init:
            self.fq = 1/(timeout/1000)
            self.header['fq(hz)'] = self.fq
            self.start_send(
                sampling_callback=self.chunk_send_json, timeout=timeout,
                on_init=on_init)

        if log:
                name_file = self.lognow(self.sens_mode)
        if test:
            t0 = time.time()
            buffer = True
        if not static:
            while True:
                try:
                    soc_data = self.soc_recv_chunk_message_json()
                    if log:
                        self.log_data_chunk_json(name_file, soc_data)
                    if buffer:
                        self.log_data_chunk_buff_json(soc_data)
                except KeyboardInterrupt:
                    if test:
                        self.time_test = abs(time.time()-t0)
                    self.stop_send()
                    print('\n')
                    self.soc.flush()
                    print('Done!')
                    break
        else:
            print(self.header_msg.format(self.sens_mode, self.header['UNIT'],
                                         self.fq))
            print('\n')
            print(('{:^15}'*len(self.header['VAR'])
                   ).format(*self.header['VAR']))
            while True:
                try:
                    soc_data = self.soc_recv_chunk_message_json()
                    self.data_print_chunk_static_json(soc_data)
                    if log:
                        self.log_data_chunk_json(name_file, soc_data)
                    if buffer:
                        self.log_data_chunk_buff_json(soc_data)
                except KeyboardInterrupt:
                    if test:
                        self.time_test = abs(time.time()-t0)
                    self.stop_send()
                    print('\n')
                    self.soc.flush()
                    print('Done!')
                    break

    # CALLBACKS

    def sample_send_call(self, x):
        pass

    def chunk_send_call(self, x):
        pass

    def chunk_send_json(self, x):
        pass

    # LOG methods

    # FILE LOG

    def lognow(self, sensor_mode, filename=None, debug=True, rtn=True):
        file_name = 'log_{}_{}.txt'.format(sensor_mode,
                                           datetime.now().strftime("%m_%d_%Y_%H_%M_%S"))
        if filename is not None:
            file_name = filename
        if debug:
            print('Saving file {} ...'.format(file_name))
        with open(file_name, 'w') as file_log:
            file_log.write(json.dumps(self.header))
            file_log.write('\n')
        if rtn:
            return file_name

    def log_data(self, filename, data):
        try:
            data_pack = dict(zip(self.header['VAR'],
                                 [val for val in data]))
            with open(filename, 'a') as file_log:
                file_log.write(json.dumps(data_pack))
                file_log.write('\n')
        except Exception as e:
            pass

    def lognow_shot(self, sensor_mode, filename=None, debug=True, rtn=True):
        file_name = filename
        if debug:
            print('Saving file {} ...'.format(file_name))
        with open(file_name, 'w') as file_log:
            header_vars = self.header['VAR'].copy()
            header_vars.append('TS')
            header_unit = self.header['UNIT']
            SHOT_HEADER = {'VAR': header_vars, 'UNIT': header_unit}
            file_log.write(json.dumps(SHOT_HEADER))
            file_log.write('\n')
        if rtn:
            return file_name

    def log_data_shot(self, filename, data, n_tag=None):
        try:
            data_pack = dict(zip(self.header['VAR']+['TS'],
                                 [val for val in data]+[datetime.now().strftime("%m_%d_%Y_%H_%M_%S")]))
            if n_tag is not None:
                data_pack = dict(zip(self.header['VAR']+['TS'],
                                     [val for val in data]+[n_tag]))
            with open(filename, 'a') as file_log:
                file_log.write(json.dumps(data_pack))
                file_log.write('\n')
        except Exception as e:
            pass

    def log_data_chunk(self, filename, data):
        try:
            data_pack = dict(zip(self.header['VAR'],
                                 [list(data)]))
            with open(filename, 'a') as file_log:
                file_log.write(json.dumps(data_pack))
                file_log.write('\n')
        except Exception as e:
            pass

    def log_data_chunk_json(self, filename, data):
        try:
            if data is not None:
                with open(filename, 'a') as file_log:
                    file_log.write(json.dumps(data))
                    file_log.write('\n')
        except Exception as e:
            pass

    # BUFFER LOG

    def log_data_buff(self, data):
        try:
            data_pack = dict(zip(self.header['VAR'],
                                 [val for val in data]))
            self.buffer.append(json.dumps(data_pack))
        except Exception as e:
            pass

    def log_data_shot_buff(self, data, n_tag=None):
        try:
            data_pack = dict(zip(self.header['VAR']+['TS'],
                                 [val for val in data]+[datetime.now().strftime("%m_%d_%Y_%H_%M_%S")]))
            if n_tag is not None:
                data_pack = dict(zip(self.header['VAR']+['TS'],
                                     [val for val in data]+[n_tag]))
            self.buffer.append(json.dumps(data_pack))
        except Exception as e:
            pass

    def log_data_chunk_buff(self, data):
        try:
            data_pack = dict(zip(self.header['VAR'],
                                 [list(data)]))
            self.buffer.append(json.dumps(data_pack))
        except Exception as e:
            pass

    def log_data_chunk_buff_json(self, data):
        try:
            if data is not None:
                self.buffer.append(json.dumps(data))
        except Exception as e:
            print(e)
            pass

    def read_buffer(self, flatten=False):
        list_of_vars = {var: [] for var in self.header['VAR']}
        for message in self.buffer:
            try:
                vals = json.loads(message)
                for key in vals.keys():
                    list_of_vars[key].append(vals[key])
            except Exception as e:
                pass
        if flatten:
            for key in vals.keys():
                list_of_vars[key] = [item for sublist in list_of_vars[key]
                                     for item in sublist]
        return (list_of_vars)

    def read_buffer_shot(self):
        list_of_vars = {var: [] for var in self.header['VAR']+['TS']}
        for message in self.buffer:
            try:
                vals = json.loads(message)
                for key in vals.keys():
                    list_of_vars[key].append(vals[key])
            except Exception as e:
                pass
        return (list_of_vars)

    def flush_buffer(self):
        self.buffer = []

    # STREAM TEST
    def get_stream_test(self, chunk=False, json=False):
        print('STREAM TEST RESULTS ARE:')
        print('TEST DURATION : {} (s)'.format(self.time_test))
        # # FIND SAMPLING RATE
        # Method 1:
        N_DATA_PACKETS = len(self.buffer)  # Number of batches received
        print('DATA PACKETS : {} packets'.format(N_DATA_PACKETS))
        # (assuming al batches = buffer_size long)
        BUFFERSIZE = self.chunk_buffer_size
        if not chunk:
            BUFFERSIZE = 1
        Fs = ((N_DATA_PACKETS+1)*BUFFERSIZE/self.time_test)
        print('SAMPLES PER PACKET : {}'.format(BUFFERSIZE))
        print('VARIABLES PER SAMPLE : {}; {}'.format(
            len(self.header['VAR']), self.header['VAR']))
        packet_size = self.chunk_buffer_data_length
        if not chunk:
            packet_size = self.data_length
        if json:
            packet_size = self.chunk_buffer_json_size
        print('SIZE OF PACKETS: {} bytes'.format(packet_size))
        print('Period: {} ms ; Fs:{} Hz, Data send rate: {} packets/s of {} samples'.format(
            round((1/Fs)*1e3), round(Fs, -1),
            round(N_DATA_PACKETS/self.time_test), BUFFERSIZE))
        kBs = round(N_DATA_PACKETS/self.time_test)*packet_size/1024
        print('DATA TRANSFER RATE (kBps): {} kB/s'.format(kBs))
        print('DATA TRANSFER RATE (Mbps): {} Mbps'.format((kBs*8)/1000))
#############################################
# SENSORS


# IMU LSM9DS1

class LSM9DS1:  # TODO: IMU_STREAMER + IMU_LOGGER
    def __init__(self, device, name='imu'):
        """Phantom LSM9DS1 class"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}

    @upy_cmd_c_r()
    def read_gyro(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def read_accel(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def read_magnet(self):
        return self.dev_dict


# WEATHER SENSOR BME280

class BME280:
    def __init__(self, device, name='bme'):
        """Phantom BME280 class"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}

    @upy_cmd_c_r()
    def read_compensated_data(self):
        return self.dev_dict


# ADC ADS1115
class ADS1115:
    def __init__(self, device, name='ads'):
        """Phantom ADS1115 class"""
        self.d = device
        self.name = name
        self.dev_dict = {'name': self.name, 'dev': device}

    @upy_cmd_c_r(rtn=False)
    def set_conv(self, rate=4, channel1=0, channel2=None):
        return self.dev_dict

    @upy_cmd_c_r()
    def raw_to_v(self, raw):
        return self.dev_dict

    @upy_cmd_c_r()
    def read(self, rate=4, channel1=0, channel2=None):
        return self.dev_dict

    @upy_cmd_c_r()
    def read_rev(self):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def alert_start(self, rate=4, channel1=0, channel2=None,
                    threshold_high=0x4000):
        return self.dev_dict

    @upy_cmd_c_r(rtn=False)
    def conversion_start(self, rate=4, channel1=0, channel2=None):
        return self.dev_dict

    @upy_cmd_c_r()
    def alert_read(self):
        return self.dev_dict

# POWER/CURRENT/VOLTAGE INA219


# IRQ SENSOR classes

# IMU (lsm9ds1)
class IMU_IRQ(IRQ_MG):
    def __init__(self, device, name, init_soc=False, port=8005, p_format='f',
                 n_vars=3, sensor=None, log_dir=None, logg=None):
        super().__init__(device, name, init_soc=init_soc, port=port,
                         p_format=p_format, n_vars=n_vars, sensor=None,
                         log_dir=log_dir, logg=logg)
        self.header = {'VAR': ['X', 'Y', 'Z'], 'UNIT': 'g=-9.8m/s^2'}
        self.sens_mode = 'ACCELEROMETER'
        self.data_print_msg = "X: {}, Y: {}, Z: {} ({})"

    @upy_cmd_c_r()
    def set_mode(self, mode):
        return self.dev_dict

    def setup_mode(self, v_mode):
        if v_mode == 'acc':
            self.header['UNIT'] = 'g=-9.8m/s^2'
            self.sens_mode = 'ACCELEROMETER'
            self.set_mode(v_mode)
        if v_mode == 'gyro':
            self.header['UNIT'] = 'deg/s'
            self.sens_mode = 'GYROSCOPE'
            self.set_mode(v_mode)
        if v_mode == 'mag':
            self.header['UNIT'] = 'gauss'
            self.sens_mode = 'MAGNETOMETER'
            self.set_mode(v_mode)


# WEATHER (bme280)

class BME_IRQ(IRQ_MG):
    def __init__(self, device, name, init_soc=False, port=8005, p_format='f',
                 n_vars=3, sensor=None, log_dir=None, logg=None):
        super().__init__(device, name, init_soc=init_soc, port=port,
                         p_format=p_format, n_vars=n_vars, sensor=None,
                         log_dir=log_dir, logg=logg)
        # CUSTOM SENS
        self.header = {'VAR': ['Temp', 'Press', 'RH'], 'UNIT': 'C ; Pa ; %'}
        self.sens_mode = 'WEATHER'
        self.header_msg = 'Streaming BME {}: Temp, Press, Rel.Hum ({}),fq={}Hz'
        self.data_print_msg = "T: {}, P: {}, RH: {} ({})"


# ADC (ADS1115)

class ADS_IRQ(IRQ_MG):
    def __init__(self, device, name, init_soc=False, port=8005, p_format='f',
                 n_vars=1, sensor=None, log_dir=None, logg=None, channel=0):
        super().__init__(device, name, init_soc=init_soc, port=port,
                         p_format=p_format, n_vars=n_vars, sensor=None,
                         log_dir=log_dir, logg=logg)
        # CUSTOM SENS
        self.channel = channel
        self.header = {'VAR': ['V'], 'UNIT': 'Volts'}
        self.sens_mode = 'ADS CHANNEl {}'.format(self.channel)
        self.header_msg = 'Streaming {}: V ({}),fq={}Hz'
        self.data_print_msg = "V: {} ({})"

    @upy_cmd_c_r()
    def set_channel(self, channel):
        return self.dev_dict

    @upy_cmd_c_r()
    def set_mode(self, mode):
        return self.dev_dict

    @upy_cmd_c_r()
    def read_shot(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def init_ads(self):
        return self.dev_dict

    def setup_channel(self, v_channel):
        self.set_channel(v_channel)
        self.channel = v_channel
        self.sens_mode = 'CHANNEl {}'.format(self.channel)
        self.init_ads()

        def init_ads_call(self):  # on_init callback
            pass
# STREAM SENSOR classes


# IMU (LSM9DS1)

class IMU_STREAMER(STREAMER):
    def __init__(self, device, name, init_soc=False, port=8005, p_format='f',
                 n_vars=3, log_dir=None, max_digit=8, chunk_buff_size=32,
                 soc_timeout=1):
        super().__init__(device, name, init_soc=init_soc, port=port,
                         p_format=p_format,
                         n_vars=n_vars, log_dir=log_dir,
                         chunk_buffer_size=chunk_buff_size,
                         soc_timeout=soc_timeout)
        # CUSTOM SENS
        self.header = {'VAR': ['X', 'Y', 'Z'], 'UNIT': 'g=-9.8m/s^2',
                       'fq(hz)': self.fq}
        self.sens_mode = 'ACCELEROMETER'
        self.header_msg = 'Streaming IMU {}: X, Y, Z ({}),fq={}Hz'
        self.data_print_msg = "X: {}, Y: {}, Z: {} ({})"
        self.max_dig = max_digit
        self.chunk_buffer_json_size = self.get_json_chunk_size()

    def get_json_chunk_size(self):
        json_size = ((self.max_dig+2)*self.chunk_buffer_size*self.n_vars)+(7*self.n_vars)
        return json_size

    @upy_cmd_c_r()
    def set_mode(self, mode):
        return self.dev_dict

    def setup_mode(self, v_mode):
        if v_mode == 'acc':
            self.header['UNIT'] = 'g=-9.8m/s^2'
            self.sens_mode = 'ACCELEROMETER'
            self.set_mode(v_mode)
        if v_mode == 'gyro':
            self.header['UNIT'] = 'deg/s'
            self.sens_mode = 'GYROSCOPE'
            self.set_mode(v_mode)
        if v_mode == 'mag':
            self.header['UNIT'] = 'gauss'
            self.sens_mode = 'MAGNETOMETER'
            self.set_mode(v_mode)

    # CALLBACK
    def chunk_send_json(self, x):
        pass


# BME280

class BME_STREAMER(STREAMER):
    def __init__(self, device, name, init_soc=False, port=8005, p_format='f',
                 n_vars=3, log_dir=None, soc_timeout=1, logg=None):
        super().__init__(device, name, init_soc=init_soc, port=port,
                         p_format=p_format,
                         n_vars=n_vars, log_dir=log_dir,
                         soc_timeout=soc_timeout, logg=logg)
        # CUSTOM SENS
        self.header = {'VAR': ['Temp', 'Press', 'RH'], 'UNIT': 'C ; Pa ; %',
                       'fq(hz)': self.fq}
        self.sens_mode = 'WEATHER'
        self.header_msg = 'Streaming BME {}: Temp, Press, Rel.Hum ({}),fq={}Hz'
        self.data_print_msg = "T: {}, P: {}, RH: {} ({})"


# ADS1115

class ADS_STREAMER(STREAMER):
    def __init__(self, device, name, init_soc=False, port=8005, p_format='f',
                 n_vars=1, log_dir=None, channel=0, soc_timeout=1):
        super().__init__(device, name, init_soc=init_soc, port=port,
                         p_format=p_format,
                         n_vars=n_vars, log_dir=log_dir,
                         soc_timeout=soc_timeout)

        # CUSTOM SENS
        self.channel = channel
        self.header = {'VAR': ['V'], 'UNIT': 'Volts',
                       'fq(hz)': self.fq}
        self.sens_mode = 'ADS CHANNEl {}'.format(self.channel)
        self.header_msg = 'Streaming {}: V ({}),fq={}Hz'
        self.data_print_msg = "V: {} ({})"

    @upy_cmd_c_r()
    def set_channel(self, channel):
        return self.dev_dict

    @upy_cmd_c_r()
    def set_mode(self, mode):
        return self.dev_dict

    @upy_cmd_c_r()
    def read_shot(self):
        return self.dev_dict

    @upy_cmd_c_r()
    def init_ads(self):
        return self.dev_dict

    def setup_channel(self, v_channel):
        self.set_channel(v_channel)
        self.channel = v_channel
        self.sens_mode = 'CHANNEl {}'.format(self.channel)
        self.init_ads()

    def init_ads_call(self):  # on_init callback
        pass

# POWER/CURRENT/VOLTAGE INA219

# UPYUTILS(OUTPUT : BUZZER, SERVO, STEPPER, DISPLAY)
#############################################
# U_LOGGER_SERVER (logging through sockets) then make a project with python logging

# SOCKETS UTILS PYTHON


# SOCKET SERVER

class socket_server:
    """
    Socket server simple class
    """

    def __init__(self, port, buff=1024, soc_timeout=1, logg=None):
        self.log = logg
        try:
            self.host = self.find_localip()
            if self.log is not None:
                self.log.info('Host IP: {}'.format(self.host))
            else:
                print(self.host)
        except Exception as e:
            if self.log is not None:
                self.log.error('Connection ERROR', exc_info=True)
            else:
                print(str(e))
            pass
        self.host_ap = '192.168.4.1'
        self.port = port
        self.serv_soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.serv_soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.addr = None
        self.buff = bytearray(buff)
        self.conn = None
        self.addr_client = None
        self.soc_timeout = soc_timeout

    def find_localip(self):
        ip_soc = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        ip_soc.connect(('8.8.8.8', 1))
        local_ip = ip_soc.getsockname()[0]
        ip_soc.close()
        return local_ip

    def start_SOC(self):
        self.serv_soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.serv_soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.serv_soc.bind((self.host, self.port))
        self.serv_soc.listen(1)
        if self.log is not None:
            self.log.info('Server listening...')
        else:
            print('Server listening...')
        self.conn, self.addr_client = self.serv_soc.accept()
        if self.log is not None:
            self.log.info(('Connection received from: {}:{}'.format(*self.addr_client)))
        else:
            print('Connection received from: {}:{}'.format(*self.addr_client))
        self.conn.settimeout(self.soc_timeout)

    def start_SOC_AP(self):
        self.serv_soc.bind((self.host_ap, self.port))
        self.serv_soc.listen(1)
        if self.log is not None:
            self.log.info('Server listening...')
        else:
            print('Server listening...')
        self.conn, self.addr_client = self.serv_soc.accept()
        if self.log is not None:
            self.log.info(('Connection received from: {}:{}'.format(*self.addr_client)))
        else:
            print('Connection received from: {}:{}'.format(*self.addr_client))
        self.conn.settimeout(self.soc_timeout)

    def flush(self):
        flushed = 0
        while flushed == 0:
            try:
                self.serv_soc.recv(128)
            except Exception as e:
                flushed = 1
                # print('flushed!')
        flushed = 0
        while flushed == 0:
            try:
                self.conn.recv(128)
            except Exception as e:
                flushed = 1
                if self.log is not None:
                    self.log.info('Flushed!')
                else:
                    print('Flushed!')

    def send_message(self, message):
        self.conn.sendall(bytes(message, 'utf-8'))

    def recv_message(self):
        self.buff[:] = self.conn.recv(len(self.buff))
        print(self.buff.decode())


# SOCKET CLIENT

class socket_client:
    """
    Socket client simple class
    """

    def __init__(self, host, port, buff=1024, soc_timeout=1):
        self.cli_soc = None
        self.host = host
        self.port = port
        self.cli_soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.cli_soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.addr = socket.getaddrinfo(self.host, self.port)[0][-1]
        self.buff = bytearray(buff)
        self.soc_timeout = soc_timeout

    def connect_SOC(self):
        self.cli_soc.connect(self.addr)
        self.cli_soc.settimeout(self.soc_timeout)

    def flush(self):
        flushed = 0
        while flushed == 0:
            try:
                self.cli_soc.recv(128)
            except Exception as e:
                flushed = 1
                print('flushed!')

    def send_message(self, message):
        self.cli_soc.sendall(bytes(message, 'utf-8'))

    def recv_message(self):
        self.buff[:] = self.cli_soc.recv(len(self.buff))
        print(self.buff.decode())

# MICROPYTHON
# U_SERVER_SOCKET
# U_SERVER_CLIENT



#!/usr/bin/env python3
# @Author: carlosgilgonzalez
# @Date:   2019-07-11T23:33:30+01:00
# @Last modified by:   carlosgilgonzalez
# @Last modified time: 2019-11-16T00:38:46+00:00

import ast
import subprocess
import shlex
import time
import serial
import struct
import multiprocessing
from dill.source import getsource
from array import array
from pexpect.replwrap import REPLWrapper
import functools


name = 'upydevice'
version = '0.1.4'


class W_UPYDEVICE:
    def __init__(self, ip_target, password, name=None, bundle_dir=''):
        self.password = password
        self.ip = ip_target
        self.response = None
        self.output = None
        self.bundle_dir = bundle_dir
        self.long_output = []
        self.process_raw = None
        self.name = name
        self.dev_class = 'WIRELESS'
        if name is None:
            self.name = 'wupydev_{}'.format(self.ip.split('.')[-1])
        self.output_queue = multiprocessing.Queue(maxsize=1)
        self._wconn = None
        self.repl_CONN = False

    def _send_recv_cmd2(self, cmd):  # test method
        resp_recv = False
        command = shlex.split(cmd)
        while not resp_recv:
            try:
                process = subprocess.Popen(command, stdout=subprocess.PIPE)
                resp_recv = True
            except Exception as e:
                pass

        stdout = process.communicate()
        try:
            resp = ast.literal_eval(
                stdout[0].decode('utf-8').split('\n')[6][4:-1])
        except Exception as e:
            try:
                resp = stdout[0].decode('utf-8').split('\n')[6][4:-1]
            except Exception as e:
                resp = None

            pass
        return resp, stdout

    def _cmd_r(self, cmd, pt=False):  # test method
        command = 'web_repl_cmd_r  -c "{}" -p {} -t {}'.format(
            cmd, self.password, self.ip)
        resp = self._send_recv_cmd2(command)
        if pt:
            print(resp[0])
        return resp[0]

    def _cmd(self, cmd):  # test method
        command = 'web_repl_cmd -c "{}" -p {} -t {}'.format(
            cmd, self.password, self.ip)
        resp = self._send_recv_cmd2(command)
        return resp[0]

    def _run_command_rl(self, command):  # test method
        end = False
        lines = []
        process = subprocess.Popen(
            shlex.split(command), stdout=subprocess.PIPE)
        while end is not True:
            if process.poll() is None:
                output = process.stdout.readline()
                if output == '' and process.poll() is not None:
                    break
                if output:
                    line = output.strip().decode()
                    lines.append(line)
                    if output.strip() == b'### closed ###':
                        end = True
            else:
                break
        rc = process.poll()
        return rc, lines

    def _cmd_rl(self, command, rt=False, evl=True):  # test method
        cmd = command
        cmd_str = 'web_repl_cmd_r -c "{}" -t {} -p {}'.format(
            cmd, self.ip, self.password)
        cmd_resp = self._run_command_rl(cmd_str)
        resp = cmd_resp[1]
        output = []
        for line in resp[6:]:
            if line == '### closed ###':
                pass
            else:
                try:
                    if line[0] == '>':
                        print(line[4:])
                        output.append(line[4:])
                    else:
                        print(line)
                        output.append(line)
                except Exception as e:
                    if len(line) == 0:
                        pass
                    else:
                        print(e)
                        pass
        if rt:
            if evl:
                return ast.literal_eval(output[0])
            else:
                return output

    def cmd(self, command, silent=False, p_queue=None, bundle_dir='', capture_output=False):  # best method
        cmd_str = self.bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format(
            command, self.ip, self.password)
        if bundle_dir is not '':
            cmd_str = bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format(
                command, self.ip, self.password)
        # print(group_cmd_str)
        self.long_output = []
        cmd = shlex.split(cmd_str)
        try:
            proc = subprocess.Popen(
                cmd, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            for i in range(6):
                proc.stdout.readline()
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if len(resp) > 0:
                    if resp[0] == '>':
                        if not silent:
                            print(resp[4:])
                        self.response = resp[4:]
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp[4:])
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp[4:])
                                except Exception as e:
                                    if 'bytearray' in resp[4:]:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp[4:]:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                    else:
                        if not silent:
                            print(resp)
                        self.response = resp
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp)
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp)
                                except Exception as e:
                                    if 'bytearray' in resp:

                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                else:
                    if not silent:
                        print(resp)

        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def cmd_p(self, command, silent=False, p_queue=None, bundle_dir='', capture_output=False):  # best method
        cmd_str = self.bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format(
            command, self.ip, self.password)
        if bundle_dir is not '':
            cmd_str = bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format(
                command, self.ip, self.password)
        # print(group_cmd_str)
        self.long_output = []
        cmd = shlex.split(cmd_str)
        try:
            proc = subprocess.Popen(
                cmd, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            for i in range(6):
                proc.stdout.readline()
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if len(resp) > 0:
                    if resp[0] == '>':
                        if not silent:
                            print('{}:{}'.format(self.name, resp[4:]))
                        self.response = resp[4:]
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp[4:])
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp[4:])
                                except Exception as e:
                                    if 'bytearray' in resp[4:]:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp[4:]:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                    else:
                        if not silent:
                            print('{}:{}'.format(self.name, resp))
                        self.response = resp
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp)
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp)
                                except Exception as e:
                                    if 'bytearray' in resp:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                else:
                    if not silent:
                        print('{}:{}'.format(self.name, resp))
        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def _cmd_nb(self, command, silent=False, time_out=2, bundle_dir=''):  # non blocking device method
        cmd_str = self.bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format(
            command, self.ip, self.password)
        if bundle_dir is not '':
            cmd_str = bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format(
                command, self.ip, self.password)
        # print(group_cmd_str)
        self.long_output = []
        cmd = shlex.split(cmd_str)
        try:
            proc = subprocess.Popen(
                cmd, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            for i in range(6):
                proc.stdout.readline()
            time.sleep(time_out)
            proc.terminate()

        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def cmd_nb(self, command, silent=False, time_out=2, block_dev=True):
        if not block_dev:
            self.dev_process_raw = multiprocessing.Process(
                target=self._cmd_nb, args=(command, silent, time_out, self.bundle_dir))
            self.dev_process_raw.start()
        else:
            self.dev_process_raw = multiprocessing.Process(
                target=self.cmd, args=(command, silent, self.output_queue, self.bundle_dir))
            self.dev_process_raw.start()

    def get_opt(self):
        try:
            self.output = self.output_queue.get(block=False)
        except Exception:
            pass

    def reset(self, bundle_dir='', output=True):
        reset_cmd_str = self.bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format('D',
                                                                                    self.ip, self.password)
        if bundle_dir is not '':
            reset_cmd_str = bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format('D',
                                                                                   self.ip, self.password)
        reset_cmd = shlex.split(reset_cmd_str)
        if output:
            print('Rebooting device...')
        try:
            proc = subprocess.Popen(
                reset_cmd, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            for i in range(6):
                proc.stdout.readline()
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if len(resp) > 0:
                    if resp[0] == '>':
                        if output:
                            print(resp[4:])
                    else:
                        if output:
                            print(resp)
                else:
                    if output:
                        print(resp)
            if output:
                print('Done!')
        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def get_output(self):
        try:
            self.output = ast.literal_eval(self.response)
        except Exception as e:
            if 'bytearray' in self.response:
                try:
                    self.output = bytearray(ast.literal_eval(
                        self.response.strip().split('bytearray')[1]))
                except Exception as e:
                    pass
            else:
                if 'array' in self.response:
                    try:
                        arr = ast.literal_eval(
                            self.response.strip().split('array')[1])
                        self.output = array(arr[0], arr[1])
                    except Exception as e:
                        pass
            pass

    def kbi(self, bundle_dir='', output=True):
        reset_cmd_str = self.bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format(hex(3),
                                                                                    self.ip, self.password)
        if bundle_dir is not '':
            reset_cmd_str = bundle_dir+'web_repl_cmd_r -c "{}" -t {} -p {}'.format(hex(3),
                                                                                   self.ip, self.password)
        reset_cmd = shlex.split(reset_cmd_str)
        if output:
            print('KeyboardInterrupt sent!')
        try:
            proc = subprocess.Popen(
                reset_cmd, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            for i in range(6):
                proc.stdout.readline()
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if len(resp) > 0:
                    if resp[0] == '>':
                        if output:
                            print(resp[4:])
                    else:
                        if output:
                            print(resp)
                else:
                    if output:
                        print(resp)
            if output:
                print('Done!')
        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def is_reachable(self, n_tries=2, max_loss=1, debug=False, timeout=2):
        ping_cmd_str = 'ping -c {} {} -t {}'.format(n_tries, self.ip, timeout)
        ping_cmd = shlex.split(ping_cmd_str)
        timeouts = 0
        down_kw = ['Unreachable', 'down', 'timeout']
        try:
            proc = subprocess.Popen(
                ping_cmd, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if debug:
                    print(resp)
                if any([kw in resp for kw in down_kw]):
                    timeouts += 1

            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                if debug:
                    print(message[:-1].decode())

        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                if debug:
                    print(message[:-1].decode())

        if timeouts >= max_loss:
            if debug:
                print('DEVICE IS DOWN OR SIGNAL RSSI IS TO LOW')
            return False
        else:
            return True

    def open_wconn(self, bundle_dir='', dbg=False):
        self._wconn = REPLWrapper(
            bundle_dir+'web_repl_conn {} -p {}'.format(self.ip, self.password), ">>> ", None)
        self.repl_CONN = True
        if dbg:
            print('WebREPL connection ready!')

    def wr_cmd(self, cmd_msg, dbg=False, silent=False, rtn=True, timeout=1, follow=True, pipe=None, m_kbi=False):
        cmd = self._wconn.child.sendline(cmd_msg)
        s_output = False
        try:
            raw_out = ' '
            cmd_echo = self._wconn.child.readline()
            timed_out = False
            len_output_now = 0
            len_output_prev = 0
            expect_prompt = ' '
            time.sleep(0.2)
            while not self._wconn.prompt.strip() in expect_prompt.strip():
                try:
                    raw_out += self._wconn.child.read_nonblocking(
                        2**16, timeout)
                    s_output = True
                    expect_prompt = raw_out.split('\n')[-1]
                    if follow:
                        outlines = [line for line in raw_out.split(
                            '...')[-1].splitlines()[:-1] if line != '']  # line.strip()
                        len_output_now = len(outlines)
                        if dbg:
                            print(outlines)
                            print('LEN NOW: {}'.format(len_output_now))
                            print('LEN PREV: {}'.format(len_output_prev))
                        if len_output_now > len_output_prev:
                            diff_len = len_output_now - len_output_prev
                            len_output_prev = len(outlines)
                            for line in outlines[-diff_len:]:
                                if line == outlines[0]:
                                    line = line[1:]
                                cmds = [val for val in cmd_msg.split(
                                    '\r') if val != '']
                                if line.replace('>>> ', '') != cmd_msg and line.replace('>>> ', '').strip() not in cmds and '\x08' not in line.replace('>>> ', ''):
                                    self.response = line.replace('>>> ', '')
                                    if pipe is not None:
                                        pipe(self.response+'\n')
                                    else:
                                        if not silent:
                                            print(self.response)
                        elif len_output_now < len_output_prev:
                            len_output_prev = 0

                except Exception as e:
                    if dbg:
                        print(e)
                        print(expect_prompt)
                    timed_out = True
                except KeyboardInterrupt:
                    if m_kbi:
                        self.close_wconn()
                        self.kbi()
                        time.sleep(0.5)
                        self.open_wconn()
                except EOFError:
                    if m_kbi:
                        self.close_wconn()
                        self.kbi()
                        time.sleep(0.5)
                        self.open_wconn()
        except Exception as e:
            if dbg:
                print('Timeout')
        if s_output:
            s_output = raw_out.replace('>>>', '').strip().split('\n')[-1]
            self.process_raw = raw_out
            outlines = [line.strip() for line in self.process_raw.split(
                '...')[-1].splitlines()[:-1] if line != '']
            for line in outlines:
                if line != cmd_msg:
                    self.response = line
                    if rtn:
                        self.get_output()
                    if not silent:
                        if not follow:
                            print(line)

    def close_wconn(self):
        self._wconn.child.close()
        self.repl_CONN = False


# S_UPYDEVICE

class S_UPYDEVICE:
    def __init__(self, serial_port, timeout=100, baudrate=9600, name=None, bundle_dir=''):
        self.serial_port = serial_port
        self.returncode = None
        self.timeout = timeout
        self.baudrate = baudrate
        self.name = name
        self.process_raw = None
        self.dev_class = 'SERIAL'
        self.bundle_dir = bundle_dir
        if name is None:
            self.name = 'supydev_{}'.format(self.serial_port.split('/')[-1])
        self.output_queue = multiprocessing.Queue(maxsize=1)
        self.picocom_cmd = shlex.split(
            'picocom -port {} -qcx {} -b{}'.format(self.serial_port, self.timeout, self.baudrate))
        self.response = None
        self.response_object = None
        self.output = None
        self.long_output = []
        self._wconn = None
        self.repl_CONN = False
        self.serial = serial.Serial(self.serial_port, self.baudrate)
        self.reset()
        # self._reset()
        self.serial.close()

    def get_output(self):
        try:
            self.output = ast.literal_eval(self.response)
        except Exception as e:
            if 'bytearray' in self.response:
                try:
                    self.output = bytearray(ast.literal_eval(
                        self.response.strip().split('bytearray')[1]))
                except Exception as e:
                    pass
            else:
                if 'array' in self.response:
                    try:
                        arr = ast.literal_eval(
                            self.response.strip().split('array')[1])
                        self.output = array(arr[0], arr[1])
                    except Exception as e:
                        pass
            pass

    def enter_cmd(self):
        if not self.serial.is_open:
            self.serial.open()
        self.serial.write(struct.pack('i', 0x0d))  # CR
        self.serial.close()

    def cmd(self, command, silent=False, p_queue=None, bundle_dir='', capture_output=False, timeout=None):
        self.long_output = []
        self.picocom_cmd = shlex.split(self.bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
            shlex.quote(command), self.timeout, self.baudrate, self.serial_port))
        if timeout is not None:
            self.picocom_cmd = shlex.split(self.bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                shlex.quote(command), timeout, self.baudrate, self.serial_port))
        if bundle_dir is not '':
            self.picocom_cmd = shlex.split(bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                shlex.quote(command), self.timeout, self.baudrate, self.serial_port))
            if timeout is not None:
                self.picocom_cmd = shlex.split(bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                    shlex.quote(command), timeout, self.baudrate, self.serial_port))
        try:
            proc = subprocess.Popen(
                self.picocom_cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            time.sleep(0.2)
            for i in range(2):
                self.enter_cmd()
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if len(resp) > 0:
                    if resp[0] == '>':
                        if not silent:
                            print(resp[4:])
                        self.response = resp[4:]
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp[4:])
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp[4:])
                                except Exception as e:
                                    if 'bytearray' in resp[4:]:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp[4:]:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                    else:
                        if resp != '{}\r'.format(command):
                            if not silent:
                                print(resp)
                        self.response = resp
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp)
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp)
                                except Exception as e:
                                    if 'bytearray' in resp:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                else:
                    if not silent:
                        print(resp)

        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def cmd_p(self, command, silent=False, p_queue=None, bundle_dir='', capture_output=False, timeout=None):
        self.long_output = []
        self.picocom_cmd = shlex.split(self.bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
            shlex.quote(command), self.timeout, self.baudrate, self.serial_port))
        if timeout is not None:
            self.picocom_cmd = shlex.split(self.bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                shlex.quote(command), timeout, self.baudrate, self.serial_port))
        if bundle_dir is not '':
            self.picocom_cmd = shlex.split(bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                shlex.quote(command), self.timeout, self.baudrate, self.serial_port))
            if timeout is not None:
                self.picocom_cmd = shlex.split(bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                    shlex.quote(command), timeout, self.baudrate, self.serial_port))
        try:
            proc = subprocess.Popen(
                self.picocom_cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            time.sleep(0.2)
            for i in range(2):
                self.enter_cmd()
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if len(resp) > 0:
                    if resp[0] == '>':
                        if not silent:
                            print('{}:{}'.format(self.name, resp[4:]))
                        self.response = resp[4:]
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp[4:])
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp[4:])
                                except Exception as e:
                                    if 'bytearray' in resp[4:]:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp[4:]:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                    else:
                        if resp != '{}\r'.format(command):
                            if not silent:
                                print('{}:{}'.format(self.name, resp))
                        self.response = resp
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp)
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp)
                                except Exception as e:
                                    if 'bytearray' in resp:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                else:
                    if not silent:
                        print('{}:{}'.format(self.name, resp))

        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def cmd_nb(self, command, silent=False):
        self.dev_process_raw = multiprocessing.Process(
            target=self.cmd, args=(command, silent, self.output_queue, self.bundle_dir))
        self.dev_process_raw.start()

    def get_opt(self):
        try:
            self.output = self.output_queue.get(block=False)
        except Exception:
            pass

    def reset(self, output=True):
        if output:
            print('Rebooting upydevice...')
        if not self.serial.is_open:
            self.serial.open()
        # time.sleep(1)
        while self.serial.inWaiting() > 0:
            self.serial.read()
        # print(self.serial.inWaiting())
        # time.sleep(1)
        self.serial.write(struct.pack('i', 0x0d))
        self.serial.write(struct.pack('i', 0x04))  # EOT
        self.serial.write(struct.pack('i', 0x0d))  # CR
        self.serial.flush()
        # print(self.serial.inWaiting())
        while self.serial.inWaiting() > 0:
            self.serial.read()
        # print(self.serial.inWaiting())
        self.serial.write(struct.pack('i', 0x0d))
        # time.sleep(1)
        self.serial.close()
        if output:
            print('Done!')

    def kbi(self, output=True):
        if output:
            print('KeyboardInterrupt sent!')
        if not self.serial.is_open:
            self.serial.open()
        # time.sleep(1)
        while self.serial.inWaiting() > 0:
            self.serial.read()
        # print(self.serial.inWaiting())
        # time.sleep(1)
        self.serial.write(struct.pack('i', 0x0d))
        self.serial.write(struct.pack('i', 0x03))  # EOT
        self.serial.write(struct.pack('i', 0x0d))  # CR
        self.serial.flush()
        # print(self.serial.inWaiting())
        while self.serial.inWaiting() > 0:
            self.serial.read()
        # print(self.serial.inWaiting())
        self.serial.write(struct.pack('i', 0x0d))
        # time.sleep(1)
        self.serial.close()
        if output:
            print('Done!')

    def open_wconn(self, bundle_dir='', dbg=False):
        self._wconn = REPLWrapper(
            bundle_dir + "picocom -t '>>> ' -b{} {}".format(self.baudrate, self.serial_port), ">>> ", None)
        cmd = self._wconn.child.sendline("\x08"*len('>>> ')+"\r")
        self.repl_CONN = True
        if dbg:
            print('Serial connection ready!')

    def wr_cmd(self, cmd_msg, dbg=False, silent=False, rtn=True, timeout=1, follow=True, pipe=None):
        if cmd_msg.endswith('\r'):
            cmd = self._wconn.child.sendline(cmd_msg)
        else:
            cmd = self._wconn.child.sendline(cmd_msg + '\r')
        s_output = False
        try:
            raw_out = ' '
            cmd_echo = self._wconn.child.readline()
            timed_out = False
            len_output_now = 0
            len_output_prev = 0
            expect_prompt = ' '
            # time.sleep(0.2)
            while not self._wconn.prompt.strip() in expect_prompt.strip():
                try:
                    raw_out += self._wconn.child.read_nonblocking(
                        2**16, timeout)
                    s_output = True
                    expect_prompt = raw_out.split('\n')[-1]
                    if follow:
                        outlines = [line for line in raw_out.split(
                            '...')[-1].splitlines()[:-1] if line != '']  # line.strip()
                        len_output_now = len(outlines)
                        if dbg:
                            print(outlines)
                            print('LEN NOW: {}'.format(len_output_now))
                            print('LEN PREV: {}'.format(len_output_prev))
                        if len_output_now > len_output_prev:
                            diff_len = len_output_now - len_output_prev
                            len_output_prev = len(outlines)
                            for line in outlines[-diff_len:]:
                                if line == outlines[0]:
                                    line = line[1:]
                                cmds = [val for val in cmd_msg.split(
                                    '\r') if val != '']
                                if line.replace('>>> ', '') != cmd_msg and line.replace('>>> ', '').strip() not in cmds and '\x08' not in line.replace('>>> ', ''):
                                    self.response = line.replace('>>> ', '')
                                    if pipe is not None:
                                        pipe(self.response+'\n')
                                    else:
                                        if not silent:
                                            print(self.response)
                        elif len_output_now < len_output_prev:
                            len_output_prev = 0

                except Exception as e:
                    if dbg:
                        print(e)
                        print(expect_prompt)
                    timed_out = True
        except Exception as e:
            if dbg:
                print('Timeout')
        if s_output:
            s_output = raw_out.replace('>>>', '').strip().split('\n')[-1]
            self.process_raw = raw_out
            outlines = [line.strip() for line in self.process_raw.split(
                '...')[-1].splitlines()[:-1] if line != '']
            for line in outlines:
                if line != cmd_msg:
                    self.response = line
                    if rtn:
                        self.get_output()
                    if not silent:
                        if not follow:
                            print(line)

    def close_wconn(self):
        self._wconn.child.close()
        self.repl_CONN = False


# PYBOARD


class PYBOARD:
    def __init__(self, serial_port, timeout=100, baudrate=9600, name=None, bundle_dir=''):
        self.serial_port = serial_port
        self.returncode = None
        self.timeout = timeout
        self.baudrate = baudrate
        self.picocom_cmd = None
        self.response = None
        self.response_object = None
        self.name = name
        self.dev_class = 'SERIAL'
        self.bundle_dir = bundle_dir
        if name is None:
            self.name = 'pyboard_{}'.format(self.serial_port.split('/')[-1])
        self.output_queue = multiprocessing.Queue(maxsize=1)
        self.output = None
        self.process_raw = None
        self.long_output = []
        self._wconn = None
        self.repl_CONN = False
        self.serial = serial.Serial(self.serial_port, self.baudrate)
        self.reset(output=False)
        self.reset(output=False)
        # self.serial.close()
        for i in range(3):
            self.enter_cmd()

    def get_output(self):
        try:
            self.output = ast.literal_eval(self.response)
        except Exception as e:
            if 'bytearray' in self.response:
                try:
                    self.output = bytearray(ast.literal_eval(
                        self.response.strip().split('bytearray')[1]))
                except Exception as e:
                    pass
            else:
                if 'array' in self.response:
                    try:
                        arr = ast.literal_eval(
                            self.response.strip().split('array')[1])
                        self.output = array(arr[0], arr[1])
                    except Exception as e:
                        pass
            pass

    def enter_cmd(self):
        if not self.serial.is_open:
            self.serial.open()
        self.serial.write(struct.pack('i', 0x0d))  # CR
        # self.serial.close()

    def cmd(self, command, silent=False, p_queue=None, bundle_dir='', out_print=True, capture_output=False, timeout=None):
        out_print = not silent
        self.long_output = []
        self.picocom_cmd = shlex.split(self.bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
            shlex.quote(command), self.timeout, self.baudrate, self.serial_port))
        if timeout is not None:
            self.picocom_cmd = shlex.split(self.bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                shlex.quote(command), timeout, self.baudrate, self.serial_port))
        if bundle_dir is not '':
            self.picocom_cmd = shlex.split(bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                shlex.quote(command), self.timeout, self.baudrate, self.serial_port))
            if timeout is not None:
                self.picocom_cmd = shlex.split(bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                    shlex.quote(command), timeout, self.baudrate, self.serial_port))
        try:
            proc = subprocess.Popen(
                self.picocom_cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            time.sleep(0.05)  # KEY FINE TUNNING
            for i in range(2):
                self.enter_cmd()
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if len(resp) > 0:
                    if resp[0] == '>':
                        if out_print:
                            print(resp[4:])
                        self.response = resp[4:]
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp[4:])
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp[4:])
                                except Exception as e:
                                    if 'bytearray' in resp[4:]:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp[4:]:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                    else:
                        if resp != '{}\r'.format(command):
                            if out_print:
                                print(resp)
                        self.response = resp
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp)
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp)
                                except Exception as e:
                                    if 'bytearray' in resp:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                else:
                    print(resp)

            while self.serial.inWaiting() > 0:
                self.serial.read()

        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def cmd_p(self, command, silent=False, p_queue=None, bundle_dir='', out_print=True, capture_output=False, timeout=None):
        out_print = not silent
        self.long_output = []
        self.picocom_cmd = shlex.split(self.bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
            shlex.quote(command), self.timeout, self.baudrate, self.serial_port))
        if timeout is not None:
            self.picocom_cmd = shlex.split(self.bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                shlex.quote(command), timeout, self.baudrate, self.serial_port))
        if bundle_dir is not '':
            self.picocom_cmd = shlex.split(bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                shlex.quote(command), self.timeout, self.baudrate, self.serial_port))
            if timeout is not None:
                self.picocom_cmd = shlex.split(bundle_dir+'picocom -t {} -qx {} -b{} {}'.format(
                    shlex.quote(command), timeout, self.baudrate, self.serial_port))
        try:
            proc = subprocess.Popen(
                self.picocom_cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                stderr=subprocess.STDOUT)
            time.sleep(0.05)  # KEY FINE TUNNING
            for i in range(2):
                self.enter_cmd()
            while proc.poll() is None:
                resp = proc.stdout.readline()[:-1].decode()
                if len(resp) > 0:
                    if resp[0] == '>':
                        if out_print:
                            print('{}:{}'.format(self.name, resp[4:]))
                        self.response = resp[4:]
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp[4:])
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp[4:])
                                except Exception as e:
                                    if 'bytearray' in resp[4:]:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp[4:]:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                    else:
                        if resp != '{}\r'.format(command):
                            if out_print:
                                print('{}:{}'.format(self.name, resp))
                        self.response = resp
                        self.get_output()
                        if capture_output:
                            self.long_output.append(resp)
                        if p_queue is not None:
                            try:
                                try:
                                    output = ast.literal_eval(resp)
                                except Exception as e:
                                    if 'bytearray' in resp:
                                        output = bytearray(ast.literal_eval(
                                            resp.strip().split('bytearray')[1]))
                                    else:
                                        if 'array' in resp:
                                            arr = ast.literal_eval(
                                                resp.strip().split('array')[1])
                                            output = array(arr[0], arr[1])
                                    pass
                                p_queue.put((
                                    output), block=False)
                            except Exception as e:
                                pass
                else:
                    print('{}:{}'.format(self.name, resp))

            while self.serial.inWaiting() > 0:
                self.serial.read()

        except KeyboardInterrupt:
            time.sleep(1)
            result = proc.stdout.readlines()
            for message in result:
                print(message[:-1].decode())

    def cmd_nb(self, command, silent=False):
        self.dev_process_raw = multiprocessing.Process(
            target=self.cmd, args=(command, silent, self.output_queue, self.bundle_dir))
        self.dev_process_raw.start()

    def get_opt(self):
        try:
            self.output = self.output_queue.get(block=False)
        except Exception:
            pass

    def reset(self, output=True):
        if output:
            print('Rebooting pyboard...')
        if not self.serial.is_open:
            self.serial.open()
        # time.sleep(1)
        while self.serial.inWaiting() > 0:
            self.serial.read()
        # print(self.serial.inWaiting())
        # time.sleep(1)
        self.serial.write(struct.pack('i', 0x0d))
        self.serial.write(struct.pack('i', 0x04))  # EOT
        self.serial.write(struct.pack('i', 0x0d))  # CR
        self.serial.flush()
        # print(self.serial.inWaiting())
        while self.serial.inWaiting() > 0:
            self.serial.read()
        # print(self.serial.inWaiting())
        self.serial.write(struct.pack('i', 0x0d))
        # time.sleep(1)
        # self.serial.close()
        if output:
            print('Done!')

    def kbi(self, output=True):
        if output:
            print('KeyboardInterrupt sent!')
        if not self.serial.is_open:
            self.serial.open()
        # time.sleep(1)
        while self.serial.inWaiting() > 0:
            self.serial.read()
        # print(self.serial.inWaiting())
        # time.sleep(1)
        self.serial.write(struct.pack('i', 0x0d))
        self.serial.write(struct.pack('i', 0x03))  # ETX
        self.serial.write(struct.pack('i', 0x0d))  # CR
        self.serial.flush()
        # print(self.serial.inWaiting())
        while self.serial.inWaiting() > 0:
            self.serial.read()
        # print(self.serial.inWaiting())
        self.serial.write(struct.pack('i', 0x0d))
        # time.sleep(1)
        # self.serial.close()
        if output:
            print('Done!')

    def open_wconn(self, bundle_dir='', dbg=False):
        self._wconn = REPLWrapper(
            bundle_dir + "picocom -t '>>> ' -b{} {}".format(self.baudrate, self.serial_port), ">>> ", None)
        cmd = self._wconn.child.sendline("\x08"*len('>>> ')+"\r")
        self.repl_CONN = True
        if dbg:
            print('Serial connection ready!')

    def wr_cmd(self, cmd_msg, dbg=False, silent=False, rtn=True, timeout=1, follow=True, pipe=None):
        if cmd_msg.endswith('\r'):
            cmd = self._wconn.child.sendline(cmd_msg)
        else:
            cmd = self._wconn.child.sendline(cmd_msg + '\r')
        s_output = False
        try:
            raw_out = ' '
            cmd_echo = self._wconn.child.readline()
            timed_out = False
            len_output_now = 0
            len_output_prev = 0
            expect_prompt = ' '
            # time.sleep(0.2)
            while not self._wconn.prompt.strip() in expect_prompt.strip():
                try:
                    raw_out += self._wconn.child.read_nonblocking(
                        2**16, timeout)
                    s_output = True
                    expect_prompt = raw_out.split('\n')[-1]
                    if follow:
                        outlines = [line for line in raw_out.split(
                            '...')[-1].splitlines()[:-1] if line != '']  # line.strip()
                        len_output_now = len(outlines)
                        if dbg:
                            print(outlines)
                            print('LEN NOW: {}'.format(len_output_now))
                            print('LEN PREV: {}'.format(len_output_prev))
                        if len_output_now > len_output_prev:
                            diff_len = len_output_now - len_output_prev
                            len_output_prev = len(outlines)
                            for line in outlines[-diff_len:]:
                                if line == outlines[0]:
                                    line = line[1:]
                                cmds = [val for val in cmd_msg.split(
                                    '\r') if val != '']
                                if line.replace('>>> ', '') != cmd_msg and line.replace('>>> ', '').strip() not in cmds and '\x08' not in line.replace('>>> ', ''):
                                    self.response = line.replace('>>> ', '')
                                    if pipe is not None:
                                        pipe(self.response+'\n')
                                    else:
                                        if not silent:
                                            print(self.response)
                        elif len_output_now < len_output_prev:
                            len_output_prev = 0

                except Exception as e:
                    if dbg:
                        print(e)
                        print(expect_prompt)
                    timed_out = True
        except Exception as e:
            if dbg:
                print('Timeout')
        if s_output:
            s_output = raw_out.replace('>>>', '').strip().split('\n')[-1]
            self.process_raw = raw_out
            outlines = [line.strip() for line in self.process_raw.split(
                '...')[-1].splitlines()[:-1] if line != '']
            for line in outlines:
                if line != cmd_msg:
                    self.response = line
                    if rtn:
                        self.get_output()
                    if not silent:
                        if not follow:
                            print(line)

    def close_wconn(self):
        self._wconn.child.close()
        self.repl_CONN = False


class GROUP:
    def __init__(self, devs=[None], name=None):
        self.name = name
        self.devs = {dev.name: dev for dev in devs}
        self.dev_process_raw_dict = None
        self.output = None
        self.output_queue = {
            dev.name: multiprocessing.Queue(maxsize=1) for dev in devs}

    def cmd(self, command, group_silent=False, dev_silent=False, ignore=[], include=[]):
        if len(include) == 0:
            include = [dev for dev in self.devs.keys()]
        for dev in ignore:
            include.remove(dev)
        for dev in include:
            if not group_silent:
                print('Sending command to {}'.format(dev))
            self.devs[dev].cmd(command, silent=dev_silent)
        self.output = {dev: self.devs[dev].output for dev in include}

    def cmd_p(self, command, group_silent=False, dev_silent=False, ignore=[], include=[], blocking=True, id=False):
        if not id:
            self.dev_process_raw_dict = {dev: multiprocessing.Process(target=self.devs[dev].cmd, args=(
                command, dev_silent, self.output_queue[dev])) for dev in self.devs.keys()}
            if len(include) == 0:
                include = [dev for dev in self.devs.keys()]
            for dev in ignore:
                include.remove(dev)
            if not group_silent:
                print('Sending command to: {}'.format(', '.join(include)))
            for dev in include:
                # self.devs[dev].cmd(command, silent=dev_silent)
                self.dev_process_raw_dict[dev].start()

            while blocking:
                dev_proc_state = [self.dev_process_raw_dict[dev].is_alive(
                ) for dev in self.dev_process_raw_dict.keys()]
                if all(state is False for state in dev_proc_state):
                    time.sleep(0.1)
                    if not group_silent:
                        print('Done!')
                    break

            try:
                self.output = {dev: self.output_queue[dev].get(
                    timeout=2) for dev in include}
            except Exception as e:
                pass
            for dev in include:
                try:
                    self.devs[dev].output = self.output[dev]
                except Exception as e:
                    pass
        else:
            self.dev_process_raw_dict = {dev: multiprocessing.Process(target=self.devs[dev].cmd_p, args=(
                command, dev_silent, self.output_queue[dev])) for dev in self.devs.keys()}
            if len(include) == 0:
                include = [dev for dev in self.devs.keys()]
            for dev in ignore:
                include.remove(dev)
            if not group_silent:
                print('Sending command to: {}'.format(', '.join(include)))
            for dev in include:
                # self.devs[dev].cmd(command, silent=dev_silent)
                self.dev_process_raw_dict[dev].start()

            while blocking:
                dev_proc_state = [self.dev_process_raw_dict[dev].is_alive(
                ) for dev in self.dev_process_raw_dict.keys()]
                if all(state is False for state in dev_proc_state):
                    time.sleep(0.1)
                    if not group_silent:
                        print('Done!')
                    break

            try:
                self.output = {dev: self.output_queue[dev].get(
                    timeout=2) for dev in include}
            except Exception as e:
                pass
            for dev in include:
                try:
                    self.devs[dev].output = self.output[dev]
                except Exception as e:
                    pass

    def get_opt(self):
        try:
            self.output = {dev: self.output_queue[dev].get(
                timeout=2) for dev in self.devs.keys()}
        except Exception as e:
            pass
        for dev in self.devs.keys():
            try:
                self.devs[dev].output = self.output[dev]
            except Exception as e:
                pass

    def reset(self, group_silent=False, output_dev=True, ignore=[], include=[]):
        if len(include) == 0:
            include = [dev for dev in self.devs.keys()]
        for dev in ignore:
            include.remove(dev)
        for dev in include:
            if not group_silent:
                print('Rebooting {}'.format(dev))
            self.devs[dev].reset(output=output_dev)


# CODE BLOCK PARSERS

# def uparser_dec(long_command):
#     lines_cmd = []
#     space_count = [0]
#     for line in long_command.split('\n')[1:]:
#         line_before = space_count[-1]
#         line_now = line.count('   ')
#         # print(line_now)
#         space_count.append(line_now)
#         if line_before > line_now:
#             if line_now > 0:
#                 lines_cmd.append(
#                     ''.join(['\b' for i in range(int((line_before/line_now)/2))]+[line.strip()]))
#                 # print('This line must be backspaced {} times: '.format((line_before/line_now)/2), line)
#             # else:
#             #     if len(line.strip()) > 0:
#             #         lines_cmd.append(''.join(['\b' for i in range(1)]+[line.strip()]))
#
#         else:
#             lines_cmd.append('\r'.join([line.strip()]))
#     return "{}\r\r".format('\r'.join(lines_cmd))

def uparser_dec(long_command, pastemode=False, end=''):
    lines_cmd = []
    space_count = [0]
    buffer_line = ''
    previous_incomplete = False
    for line in long_command.split('\n')[1:]:
        line_before = space_count[-1]
        if line != '':
            if not previous_incomplete:
                line_now = line.count('    ')
                # print(line_now)
        # print(line_now)
            space_count.append(line_now)
            if line_before > line_now:
                if line_now > 0:
                    lines_cmd.append(
                        ''.join(['\b' for i in range(int("{:.0f}".format((line_before-line_now))))]+[line.strip()]))
                    # print('This line must be backspaced {:.0f} times: {}'.format(((line_before-line_now)), line.strip()))
                # else:
                #     if len(line.strip()) > 0:
                #         lines_cmd.append(''.join(['\b' for i in range(1)]+[line.strip()]))

            elif line[-1] == ',':
                # print('line incomplete')
                previous_incomplete = True
                buffer_line += line.strip()
            else:
                if buffer_line != '':
                    if previous_incomplete:
                        # print('This is the complete line:{}'.format(buffer_line+line.strip()))
                        lines_cmd.append('\r'.join([buffer_line+line.strip()]))
                        buffer_line = ''
                        previous_incomplete = False
                else:
                    lines_cmd.append('\r'.join([line.strip()]))
    if not pastemode:
        return "{}{}{}".format('\r'.join(lines_cmd), '\r'*line_now, end)
    else:
        return "\x05{}\x04".format(long_command)


def upy_code(func):  # TODO: ACCEPT DEVICE ARG
    def wrapper_get_str_func(*args, **kwargs):
        print(getsource(func))
        return uparser_dec(getsource(func))
    return wrapper_get_str_func


# PYTHON PHANTOM DECORATORS

def upy_cmd(device, debug=False, rtn=True):
    def decorator_cmd_str(func):
        @functools.wraps(func)
        def wrapper_cmd(*args, **kwargs):
            args_repr = [repr(a) for a in args]
            kwargs_repr = [f"{k}={v!r}" if not callable(
                v) else f"{k}={v.__name__}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            cmd = f"{func.__name__}({signature})"
            device.output = None
            if debug:
                device.cmd(cmd)
            else:
                device.cmd(cmd, silent=True)
            if rtn:
                return device.output
            else:
                return None
        return wrapper_cmd
    return decorator_cmd_str


def upy_cmd_c(device, debug=False, rtn=True, out=False):
    def decorator_cmd_str(func):
        @functools.wraps(func)
        def wrapper_cmd(*args, **kwargs):
            flags = ['>', '<', 'object', 'at', '0x']
            args_repr = [repr(a) for a in args if any(
                f not in repr(a) for f in flags)]
            kwargs_repr = [f"{k}={v!r}" if not callable(
                v) else f"{k}={v.__name__}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            cmd_ = f"{func.__name__}({signature})"
            name = func(*args, **kwargs)
            cmd = "{}.{}".format(name, cmd_)
            device.output = None
            if out:
                cmd = "{}".format(cmd_)
            else:
                pass
            if debug:
                device.cmd(cmd)
            else:
                device.cmd(cmd, silent=True)
            if rtn:
                return device.output
            else:
                return None
        return wrapper_cmd
    return decorator_cmd_str


def upy_cmd_c_raw(device, out=False):
    def decorator_cmd_str(func):
        @functools.wraps(func)
        def wrapper_cmd(*args, **kwargs):
            flags = ['>', '<', 'object', 'at', '0x']
            args_repr = [repr(a) for a in args if any(
                f not in repr(a) for f in flags)]
            kwargs_repr = [f"{k}={v!r}" if not callable(
                v) else f"{k}={v.__name__}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            cmd_ = f"{func.__name__}({signature})"
            name = func(*args, **kwargs)
            cmd = "{}.{}".format(name, cmd_)
            device.output = None
            if out:
                cmd = "{}".format(cmd_)
            else:
                pass
            device.cmd(cmd, capture_output=True)
            try:
                device.output = device.long_output[0].strip()
            except Exception as e:
                print(e)
                pass
            return None
        return wrapper_cmd
    return decorator_cmd_str


def upy_cmd_c_r(debug=False, rtn=True, out=False):
    def decorator_cmd_str(func):
        @functools.wraps(func)
        def wrapper_cmd(*args, **kwargs):
            flags = ['>', '<', 'object', 'at', '0x']
            args_repr = [repr(a) for a in args if any(
                f not in repr(a) for f in flags)]
            kwargs_repr = [f"{k}={v!r}" if not callable(
                v) else f"{k}={v.__name__}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            cmd_ = f"{func.__name__}({signature})"
            dev_dict = func(*args, **kwargs)
            cmd = "{}.{}".format(dev_dict['name'], cmd_)
            dev_dict['dev'].output = None
            if out:
                cmd = "{}".format(cmd_)
            else:
                pass
            if debug:
                dev_dict['dev'].cmd(cmd)
            else:
                dev_dict['dev'].cmd(cmd, silent=True)
            if rtn:
                return dev_dict['dev'].output
            else:
                return None
        return wrapper_cmd
    return decorator_cmd_str


def upy_cmd_c_raw_r(out=False):
    def decorator_cmd_str(func):
        @functools.wraps(func)
        def wrapper_cmd(*args, **kwargs):
            flags = ['>', '<', 'object', 'at', '0x']
            args_repr = [repr(a) for a in args if any(
                f not in repr(a) for f in flags)]
            kwargs_repr = [f"{k}={v!r}" if not callable(
                v) else f"{k}={v.__name__}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            cmd_ = f"{func.__name__}({signature})"
            dev_dict = func(*args, **kwargs)
            cmd = "{}.{}".format(dev_dict['name'], cmd_)
            dev_dict['dev'].output = None
            if out:
                cmd = "{}".format(cmd_)
            else:
                pass
            dev_dict['dev'].cmd(cmd, capture_output=True)
            try:
                dev_dict['dev'].output = dev_dict['dev'].long_output[0].strip()
            except Exception as e:
                print(e)
                pass
            return None
        return wrapper_cmd
    return decorator_cmd_str


def upy_cmd_c_r_in_callback(debug=False, rtn=True, out=False):
    def decorator_cmd_str(func):
        @functools.wraps(func)
        def wrapper_cmd(*args, **kwargs):
            flags = ['>', '<', 'object', 'at', '0x']
            args_repr = [repr(a) for a in args if any(
                f not in repr(a) for f in flags)]
            dev_dict = func(*args, **kwargs)
            name = dev_dict['name']
            kwargs_repr = [f"{k}={v!r}" if not callable(
                v) else f"{k}={name}.{v.__name__}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            cmd_ = f"{func.__name__}({signature})"
            cmd = "{}.{}".format(dev_dict['name'], cmd_)
            dev_dict['dev'].output = None
            if out:
                cmd = "{}".format(cmd_)
            else:
                pass
            if debug:
                dev_dict['dev'].cmd(cmd)
            else:
                dev_dict['dev'].cmd(cmd, silent=True)
            if rtn:
                return dev_dict['dev'].output
            else:
                return None
        return wrapper_cmd
    return decorator_cmd_str


def upy_cmd_c_r_nb(debug=False, rtn=True, out=False):
    def decorator_cmd_str(func):
        @functools.wraps(func)
        def wrapper_cmd(*args, **kwargs):
            flags = ['>', '<', 'object', 'at', '0x']
            args_repr = [repr(a) for a in args if any(
                f not in repr(a) for f in flags)]
            kwargs_repr = [f"{k}={v!r}" if not callable(
                v) else f"{k}={v.__name__}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            cmd_ = f"{func.__name__}({signature})"
            dev_dict = func(*args, **kwargs)
            cmd = "{}.{}".format(dev_dict['name'], cmd_)
            dev_dict['dev'].output = None
            if out:
                cmd = "{}".format(cmd_)
            else:
                pass
            if debug:
                dev_dict['dev'].cmd_nb(cmd)
            else:
                dev_dict['dev'].cmd_nb(cmd, silent=True)
            if rtn:
                dev_dict['dev'].get_opt()
                return dev_dict['dev'].output
            else:
                return None
        return wrapper_cmd
    return decorator_cmd_str


def upy_cmd_c_r_nb_in_callback(debug=False, rtn=True, out=False):
    def decorator_cmd_str(func):
        @functools.wraps(func)
        def wrapper_cmd(*args, **kwargs):
            flags = ['>', '<', 'object', 'at', '0x']
            args_repr = [repr(a) for a in args if any(
                f not in repr(a) for f in flags)]
            dev_dict = func(*args, **kwargs)
            name = dev_dict['name']
            kwargs_repr = [f"{k}={v!r}" if not callable(
                v) else f"{k}={name}.{v.__name__}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            cmd_ = f"{func.__name__}({signature})"
            cmd = "{}.{}".format(dev_dict['name'], cmd_)
            dev_dict['dev'].output = None
            if out:
                cmd = "{}".format(cmd_)
            else:
                pass
            if debug:
                dev_dict['dev'].cmd_nb(cmd)
            else:
                dev_dict['dev'].cmd_nb(cmd, silent=True)
            if rtn:
                dev_dict['dev'].get_opt()
                return dev_dict['dev'].output
            else:
                return None
        return wrapper_cmd
    return decorator_cmd_str


# LOAD DEVICE CONFIGURATION FUNCTIONS (json) devtools.py (submodule)

# DEFAULT IS GLOBAL DIR, BUT dir= option available (in case of bundle_dir)
# MAKE A GLOBAL DIR TO STORE DEV CONFIGURATIONS, (.upydevice_devs in $HOME)
# load_dev(name, dir=), returns {'ip':X, 'passwd':X}
# store_dev(name, ip=None, pass=None, s_port=None, dir=$HOME/.upydevice_devs)
# if not .upydevice_devs then create
# if ip:
# json dumps {}
# else:
# if s_port:
# json dumps {}
# dev = load_dev('my_esp32')
# esp32 = W_UPYDEVICE(dev['ip'], dev['passwd'])
# dev = load_dev('my_pyb')
# pyboard = PYBOARD(dev['s_port'])



#!/usr/bin/env python 3
# @Author: carlosgilgonzalez
# @Date:   2019-11-16T00:31:56+00:00
# @Last modified by:   carlosgilgonzalez
# @Last modified time: 2019-11-16T01:26:58+00:00

"""upydevice dev tools"""

import os
import json

dev_dir = '.upydevices'
dev_path = "{}/{}".format(os.environ['HOME'], dev_dir)


def setup_devs_dir(debug=False, name_dir=dev_dir):
    if name_dir not in os.listdir("{}".format(os.environ['HOME'])):
        os.mkdir("{}/{}".format(os.environ['HOME'], name_dir))
        if debug:
            print('{} directory created in HOME:{} directory'.format(name_dir, os.environ['HOME']))


def store_dev(name, ip=None, passwd=None, s_port=None, dir=dev_path,
              debug=False, **kargs):
    dev_ip = ip
    dev_pass = passwd
    dev_s_port = s_port
    dev_conf = {'ip': dev_ip, 'passwd': dev_pass,
                's_port': dev_s_port}
    file_conf = '{}/{}.config'.format(dir, name)
    for key, val in kargs.items():
        if key not in dev_conf.keys():
            dev_conf[key] = val
    with open(file_conf, 'w') as config_file:
        config_file.write(json.dumps(dev_conf))
    if debug:
        print('device {} settings saved in {} directory!'.format(name, dir))


def load_dev(name, dir=dev_path, debug=False):
    try:
        file_conf = '{}/{}.config'.format(dir, name)
        with open(file_conf, 'r') as config_file:
            dev_conf = json.loads(config_file.read())
        return dev_conf
    except Exception as e:
        if debug:
            print("CONFIGURATION FILE NOT FOUND")
        return None
